__title__ = "mathy_envs"
__version__ = "0.12.1"
__summary__ = "Learning environments for solving math problems step-by-step"
__uri__ = "https://mathy.ai"
__author__ = "Justin DuJardin"
__email__ = "justin@dujardinconsulting.com"
__license__ = "All rights reserved"



import sys
from typing import Any, cast

import mock
import numpy as np
import pytest

from mathy_envs.env import MathyEnv
from mathy_envs.state import MathyObservation


def test_gym_raises_helpful_error_if_gym_is_not_installed():
    with mock.patch.dict(sys.modules, {"gymnasium": None}):
        with pytest.raises(ImportError, match=r"pip install mathy_envs\[gym\]"):
            import mathy_envs.gym  # noqa


def test_gym_instantiate_envs():
    import gymnasium as gym

    from mathy_envs.gym import MathyGymEnv

    all_envs = gym.registry.values()
    # Filter to just mathy registered envs
    mathy_gym_envs = [e for e in all_envs if e.id.startswith("mathy-")]

    assert len(mathy_gym_envs) > 0

    # Each env can be created and produce an initial observation without
    # special configuration.
    for gym_env_spec in mathy_gym_envs:
        print(f"Testing {gym_env_spec.id}...")
        wrapper_env: MathyGymEnv = gym.make(gym_env_spec.id)  # type:ignore
        assert wrapper_env is not None

        obs = wrapper_env.reset()
        print("initial observation:", obs)
        action = wrapper_env.action_space.sample()
        wrapper_env.step(action)
        observation, info = wrapper_env.reset()
        assert isinstance(observation, (dict, np.ndarray, MathyObservation))
        assert observation is not None


def test_gym_step_info():
    import gymnasium as gym

    from mathy_envs.gym import MathyGymEnv

    all_envs = gym.registry.values()
    # Filter to just mathy registered envs
    mathy_gym_envs = [e for e in all_envs if e.id.startswith("mathy-")]

    assert len(mathy_gym_envs) > 0

    # Each env can be created and produce an initial observation without
    # special configuration.
    for gym_env_spec in mathy_gym_envs:
        print(f"Testing {gym_env_spec.id}...")
        wrapper_env: MathyGymEnv = gym.make(gym_env_spec.id)  # type:ignore
        assert wrapper_env is not None

        obs = wrapper_env.reset()
        print("initial observation:", obs)
        action = wrapper_env.action_space.sample()
        obs, reward, terminated, truncated, info = wrapper_env.step(action)
        for key in ["valid", "done", "truncated", "transition"]:
            assert key in info, f"info should have a string key {key}"


def test_gym_env_spaces():
    import gymnasium as gym

    from mathy_envs.gym import MaskedDiscrete, MathyGymEnv

    wrapper_env: MathyGymEnv = gym.make("mathy-poly-easy-v0")  # type:ignore
    mathy: MathyEnv = wrapper_env.mathy
    observation, info = wrapper_env.reset()

    # Has a masked discrete (finite) action space
    assert wrapper_env.action_space is not None
    assert isinstance(wrapper_env.action_space, MaskedDiscrete)
    # Action space is (num_rules * max_seq_len)
    assert wrapper_env.action_size == int(len(mathy.rules) * mathy.max_seq_len)
    assert wrapper_env.action_space.shape == tuple()

    # Observation matches the space spec
    assert wrapper_env.observation_space.shape == cast(Any, observation).shape


def test_gym_probability_action_mask():
    import gymnasium as gym

    from mathy_envs.gym import MathyGymEnv

    env: MathyGymEnv = gym.make(
        "mathy-poly-easy-v0", mask_as_probabilities=True
    )  # type:ignore
    obs, info = env.reset()
    offset = -env.mathy.action_size
    action_mask = np.array(obs[offset:])
    # When returned as probabilities, the mask sums to almost 1.0
    assert np.sum(action_mask) > 1.0 - 1e-4

    env: MathyGymEnv = gym.make(
        "mathy-poly-easy-v0", mask_as_probabilities=False
    )  # type:ignore
    obs, info = env.reset()
    # When not as probabilities, the mask is a bunch of 1s and 0s
    action_mask = np.array(obs[offset:])
    assert np.sum(action_mask) > 1



from typing import Any

from ..envs.poly_simplify_blockers import PolySimplifyBlockers
from ..types import MathyEnvDifficulty, MathyEnvProblemArgs
from .mathy_gym_env import MathyGymEnv, safe_register

#
# Commute + simplify with blockers
#


class GymPolynomialBlockers(MathyGymEnv):
    def __init__(self, difficulty: MathyEnvDifficulty, **kwargs: Any):
        super(GymPolynomialBlockers, self).__init__(
            env_class=PolySimplifyBlockers,
            env_problem_args=MathyEnvProblemArgs(difficulty=difficulty),
            **kwargs
        )


class PolynomialBlockersEasy(GymPolynomialBlockers):
    def __init__(self, **kwargs: Any):
        super(PolynomialBlockersEasy, self).__init__(
            difficulty=MathyEnvDifficulty.easy, **kwargs
        )


class PolynomialBlockersNormal(GymPolynomialBlockers):
    def __init__(self, **kwargs: Any):
        super(PolynomialBlockersNormal, self).__init__(
            difficulty=MathyEnvDifficulty.normal, **kwargs
        )


class PolynomialBlockersHard(GymPolynomialBlockers):
    def __init__(self, **kwargs: Any):
        super(PolynomialBlockersHard, self).__init__(
            difficulty=MathyEnvDifficulty.hard, **kwargs
        )


safe_register(
    id="mathy-poly-blockers-easy-v0",
    entry_point="mathy_envs.gym:PolynomialBlockersEasy",
)
safe_register(
    id="mathy-poly-blockers-normal-v0",
    entry_point="mathy_envs.gym:PolynomialBlockersNormal",
)
safe_register(
    id="mathy-poly-blockers-hard-v0",
    entry_point="mathy_envs.gym:PolynomialBlockersHard",
)



from typing import List, Tuple

import numpy as np

from mathy_envs.envs.poly_simplify import PolySimplify
from mathy_envs.state import MathyEnvState, MathyObservation


def test_state_to_observation():
    """to_observation has defaults to allow calling with no arguments"""
    env_state = MathyEnvState(problem="4x+2")
    assert env_state.to_observation() is not None


def test_state_to_observation_normalization():
    """normalize argument converts all values to range 0.0-1.0"""
    env_state = MathyEnvState(problem="4+2")
    obs: MathyObservation = env_state.to_observation(normalize=False)
    assert np.max(obs.values) == 4.0

    norm: MathyObservation = env_state.to_observation(normalize=True)
    assert np.max(norm.values) == 1.0


def test_state_to_observation_normalized_problem_type():
    """normalize argument converts all values and type hash to range 0.0-1.0"""
    env_state = MathyEnvState(problem="4+2")
    obs: MathyObservation = env_state.to_observation()
    print(obs.type)
    assert np.max(obs.time) <= 1.0
    assert np.min(obs.time) >= 0.0

    assert np.max(obs.values) <= 1.0
    assert np.min(obs.values) >= 0.0

    assert np.max(obs.type) <= 1.0
    assert np.min(obs.type) >= 0.0


def test_state_encodes_hierarchy():
    """Verify that the observation generated encodes hierarchy properly
    so the model can determine the precise nodes to act on"""

    diff_pairs: List[Tuple[str, str]] = [
        ("4x + (3u + 7x + 3u) + 4u", "4x + 3u + 7x + 3u + 4u"),
        ("7c * 5", "7 * (c * 5)"),
        ("5v + 20b + (10v + 7b)", "5v + 20b + 10v + 7b"),
        ("5s + 60 + 12s + s^2", "5s + 60 + (12s + s^2)"),
    ]
    env = PolySimplify()

    for one, two in diff_pairs:
        state_one = MathyEnvState(problem=one)
        obs_one = state_one.to_observation(env.get_valid_moves(state_one))

        state_two = MathyEnvState(problem=two)
        obs_two = state_two.to_observation(env.get_valid_moves(state_two))

        assert obs_one.nodes != obs_two.nodes


def test_state_sanity():
    state = MathyEnvState(problem="4+4")
    assert state is not None


def test_state_encode_player():
    env_state = MathyEnvState(problem="4x+2")
    env_state = env_state.get_out_state(
        problem="2+4x", moves_remaining=10, action=(0, 0)
    )
    agent = env_state.agent
    assert agent.problem == "2+4x"
    assert agent.moves_remaining == 10
    assert agent.action == (0, 0)


def test_state_serialize_string():
    env_state = MathyEnvState(problem="4x+2")
    for i in range(10):
        env_state = env_state.get_out_state(
            problem="2+4x", moves_remaining=10 - i, action=(i, i)
        )

    state_str = env_state.to_string()
    compare = MathyEnvState.from_string(state_str)
    assert env_state.agent.problem == compare.agent.problem
    assert env_state.agent.moves_remaining == compare.agent.moves_remaining
    for one, two in zip(env_state.agent.history, compare.agent.history):
        assert one.raw == two.raw
        assert one.action == two.action


def test_state_serialize_numpy():
    env_state = MathyEnvState(problem="4x+2")
    for i in range(10):
        env_state = env_state.get_out_state(
            problem="2+4x", moves_remaining=10 - i, action=(i, i)
        )

    state_np = env_state.to_np()
    compare = MathyEnvState.from_np(state_np)
    assert env_state.agent.problem == compare.agent.problem
    assert env_state.agent.moves_remaining == compare.agent.moves_remaining
    for one, two in zip(env_state.agent.history, compare.agent.history):
        assert one.raw == two.raw
        assert one.action == two.action



from mathy_core.problems import gen_combine_terms_in_place, rand_bool

from ..env import MathyEnvProblem
from ..types import MathyEnvDifficulty, MathyEnvProblemArgs
from .poly_simplify import PolySimplify


class PolyCombineInPlace(PolySimplify):
    """A Mathy environment for combining like terms in-place without
    any commuting. This task is intended to test the model's ability
    to identify like-terms among a bunch of unlike terms and combine
    them with a sequence of two moves.
    """

    def get_env_namespace(self) -> str:
        return "mathy.polynomials.combine.in.place"

    def max_moves_fn(
        self, problem: MathyEnvProblem, config: MathyEnvProblemArgs
    ) -> int:
        """When combining terms that are already siblings, we only need
        to take two actions:

            1. distributive factor out the common element
            2. simplify the remaining constants

        """

        return 2

    def problem_fn(self, params: MathyEnvProblemArgs) -> MathyEnvProblem:
        if params.difficulty == MathyEnvDifficulty.easy:
            powers = rand_bool(20)
            easy = rand_bool(65)
            text, _ = gen_combine_terms_in_place(
                min_terms=6, max_terms=12, easy=easy, powers=powers
            )
        elif params.difficulty == MathyEnvDifficulty.normal:
            powers = rand_bool(40)
            easy = rand_bool(15)
            text, _ = gen_combine_terms_in_place(
                min_terms=9, max_terms=16, easy=easy, powers=powers
            )
        elif params.difficulty == MathyEnvDifficulty.hard:
            powers = rand_bool(60)
            easy = rand_bool(5)
            text, _ = gen_combine_terms_in_place(
                min_terms=16, max_terms=22, easy=easy, powers=powers
            )
        else:
            raise ValueError(f"Unknown difficulty: {params.difficulty}")
        return MathyEnvProblem(text, 2, self.get_env_namespace())



from pathlib import Path
from subprocess import check_output
from ruamel.yaml import YAML

# Constants
EXCLUDE_FILES = {
    ".DS_Store",
    "__init__.py",
    "README.md",
    "types.py",
    "cli.py",
    "py.typed",
    "about.py",
    "conrastive.py",
}
INCLUDE_FOLDERS = [".", "envs"]
DOCS_KEY = "API"

# Paths
parent_folder_path = Path(__file__).parent.parent
yaml_path = parent_folder_path / "mkdocs.yml"
source_path = parent_folder_path.parent / "mathy_envs"
dest_dir = parent_folder_path / "docs" / "api"

# YAML
yaml = YAML()


def convert_module_to_title(input_string: str) -> str:
    return input_string.replace("_", " ").title()


def prepend_md_content(original_md, prepending_md):
    with open(prepending_md, "r") as file:
        prepending_content = file.read()

    with open(original_md, "r+") as file:
        original_content = file.readlines()
        # First line is a heading 1 with the full module name that is described.
        # Make it explode if that changes so we notice.
        assert original_content[0].startswith(
            "# "
        ), "Expected heading 1 at beginning of mathy_pydoc API doc file"

        # Change to
        ticks = "```"
        import_statement = f"{ticks}python\n\nimport {original_content[0][2:]}{ticks}"
        new_content = f"## API\n\n"
        module_namespace = import_statement
        original_content[0] = new_content
        # flatten the list of lines into a single string
        original_content = "".join(original_content)

        file.seek(0, 0)
        file.write(f"{module_namespace}\n{prepending_content}\n\n{original_content}")


def h1_to_h2(original_md: str):
    file_path = Path(original_md)
    original_content = file_path.read_text().splitlines()
    # First line is a heading 1 with the full module name that is described.
    # Make it explode if that changes so we notice.
    assert original_content[0].startswith(
        "# "
    ), "Expected heading 1 at beginning of mathy_pydoc API doc file"
    ticks = "```"
    original_content[0] = f"{ticks}python\n\nimport {original_content[0][2:]}\n{ticks}"
    # flatten the list of lines into a single string
    original_content = "\n".join(original_content)
    file_path.unlink()
    file_path.write_text(original_content)


def render_docs(src_rel_path, src_file, to_file, modifier="++"):
    insert = "." + src_rel_path if src_rel_path not in ["", "."] else ""
    namespace = f"mathy_envs{insert}.{src_file.stem}{modifier}"
    args = ["mathy_pydoc", "--plain", namespace]
    if not to_file.parent.exists():
        to_file.parent.mkdir(parents=True)
    call_result = check_output(args, cwd=parent_folder_path).decode("utf-8")
    with open(to_file, "w") as file:
        file.write(call_result)


def process_directory(directory):
    nav_entries = []
    for file_path in directory.iterdir():
        if file_path.name in EXCLUDE_FILES or not file_path.suffix == ".py":
            continue

        print(f"\t{file_path.name}")
        rel_out_md = file_path.with_suffix(".md").name
        to_file = dest_dir / directory.relative_to(source_path) / rel_out_md
        render_docs(directory.relative_to(source_path).as_posix(), file_path, to_file)

        # Prepend existing md file content if present
        existing_md = file_path.with_suffix(".md")
        if existing_md.exists():
            prepend_md_content(to_file, existing_md)
        else:
            h1_to_h2(to_file)

        nav_item = {
            convert_module_to_title(file_path.stem): to_file.relative_to(
                parent_folder_path / "docs"
            ).as_posix()
        }
        nav_entries.append(nav_item)
    return nav_entries


def update_yaml_nav(nav_entries):
    mkdocs_yaml = yaml.load(yaml_path)
    updated = False
    site_nav = mkdocs_yaml["nav"]
    for nav_obj in site_nav:
        site_keys = list(nav_obj.keys())
        for key in site_keys:
            if isinstance(nav_obj[key], str):
                continue
            for nav_sub in nav_obj[key]:
                if DOCS_KEY in nav_sub:
                    nav_sub[DOCS_KEY] = nav_entries
                    updated = True
                    break
    if not updated:
        raise Exception(f"Could not find {DOCS_KEY} in mkdocs.yml")
    with open(yaml_path, "w") as file:
        yaml.dump(mkdocs_yaml, file)


def main():
    print("Building API docs...")
    nav_entries = []
    for src_folder in INCLUDE_FOLDERS:
        folder = source_path / src_folder
        if folder.is_dir():
            print(f"Found directory: {folder.relative_to(source_path)}")
            if src_folder not in [".", ""]:
                new_entries = process_directory(folder)
                new_entries.sort(key=lambda x: list(x)[0])
                nav_entries.append({convert_module_to_title(folder.name): new_entries})
            else:
                nav_entries += process_directory(folder)
    nav_entries.sort(key=lambda x: list(x)[0])
    update_yaml_nav(nav_entries)
    print("Done!")


if __name__ == "__main__":
    main()



from typing import Optional

from mathy_core.expressions import MathExpression
from mathy_core.problems import (
    gen_binomial_times_binomial,
    gen_binomial_times_monomial,
    rand_bool,
)
from mathy_core.util import get_terms, has_like_terms, is_preferred_term_form
from numpy.random import uniform

from .. import time_step
from ..env import MathyEnv, MathyEnvProblem
from ..state import MathyEnvState, MathyObservation
from ..types import MathyEnvDifficulty, MathyEnvProblemArgs


class BinomialDistribute(MathyEnv):
    """A Mathy environment for distributing pairs of binomials.

    The FOIL method is sometimes used to solve these types of problems, where
    FOIL is just the distributive property applied to two binomials connected
    with a multiplication."""

    def get_env_namespace(self) -> str:
        return "mathy.binomials.mulptiply"

    def max_moves_fn(
        self, problem: MathyEnvProblem, config: MathyEnvProblemArgs
    ) -> int:
        return problem.complexity * 4

    def transition_fn(
        self,
        env_state: MathyEnvState,
        expression: MathExpression,
        features: MathyObservation,
    ) -> Optional[time_step.TimeStep]:
        """If there are no like terms."""
        if not has_like_terms(expression):
            term_nodes = get_terms(expression)
            is_win = True
            for term in term_nodes:
                if not is_preferred_term_form(term):
                    is_win = False
            if is_win:
                return time_step.termination(features, self.get_win_signal(env_state))
        return None

    def problem_fn(self, params: MathyEnvProblemArgs) -> MathyEnvProblem:
        """Given a set of parameters to control term generation, produce
        2 binomials expressions connected by a multiplication."""
        if params.difficulty == MathyEnvDifficulty.easy:
            if rand_bool(50):
                text, complexity = gen_binomial_times_monomial(min_vars=2, max_vars=3)
            else:
                text, complexity = gen_binomial_times_binomial(
                    min_vars=2,
                    max_vars=3,
                    powers_probability=uniform(0.1, 0.4),
                    like_variables_probability=uniform(0.3, 0.7),
                )
        elif params.difficulty == MathyEnvDifficulty.normal:
            text, complexity = gen_binomial_times_binomial(
                min_vars=2,
                max_vars=2,
                powers_probability=uniform(0.2, 0.6),
                like_variables_probability=uniform(0.2, 0.5),
            )
        elif params.difficulty == MathyEnvDifficulty.hard:
            text, complexity = gen_binomial_times_binomial(
                min_vars=2,
                max_vars=3,
                simple_variables=False,
                powers_probability=uniform(0.4, 0.8),
                like_variables_probability=uniform(0.1, 0.3),
            )
            complexity += 2
        else:
            raise ValueError(f"Unknown difficulty: {params.difficulty}")
        return MathyEnvProblem(text, complexity, self.get_env_namespace())



from typing import Any

from ..envs.poly_grouping import PolyGroupLikeTerms
from ..types import MathyEnvDifficulty, MathyEnvProblemArgs
from .mathy_gym_env import MathyGymEnv, safe_register

#
# Group like terms
#


class GymPolynomialGrouping(MathyGymEnv):
    def __init__(self, difficulty: MathyEnvDifficulty, **kwargs: Any):
        super(GymPolynomialGrouping, self).__init__(
            env_class=PolyGroupLikeTerms,
            env_problem_args=MathyEnvProblemArgs(difficulty=difficulty),
            **kwargs
        )


class PolynomialGroupingEasy(GymPolynomialGrouping):
    def __init__(self, **kwargs: Any):
        super(PolynomialGroupingEasy, self).__init__(
            difficulty=MathyEnvDifficulty.easy, **kwargs
        )


class PolynomialGroupingNormal(GymPolynomialGrouping):
    def __init__(self, **kwargs: Any):
        super(PolynomialGroupingNormal, self).__init__(
            difficulty=MathyEnvDifficulty.normal, **kwargs
        )


class PolynomialGroupingHard(GymPolynomialGrouping):
    def __init__(self, **kwargs: Any):
        super(PolynomialGroupingHard, self).__init__(
            difficulty=MathyEnvDifficulty.hard, **kwargs
        )


safe_register(
    id="mathy-poly-grouping-easy-v0",
    entry_point="mathy_envs.gym:PolynomialGroupingEasy",
)
safe_register(
    id="mathy-poly-grouping-normal-v0",
    entry_point="mathy_envs.gym:PolynomialGroupingNormal",
)
safe_register(
    id="mathy-poly-grouping-hard-v0",
    entry_point="mathy_envs.gym:PolynomialGroupingHard",
)



def test_envs_overview_custom_problem_text():
    from ...docs.snippets.envs import custom_problem_text  # noqa


def test_envs_overview_custom_win_conditions():
    from ...docs.snippets.envs import custom_win_conditions  # noqa


def test_envs_overview_custom_timestep_rewards():
    from ...docs.snippets.envs import custom_timestep_rewards  # noqa


def test_envs_overview_custom_episode_rewards():
    from ...docs.snippets.envs import custom_episode_rewards  # noqa


def test_envs_overview_custom_actions():
    from ...docs.snippets.envs import custom_actions  # noqa


def test_envs_overview_openai_gym():
    from ...docs.snippets.envs import openai_gym  # noqa


def test_envs_overview_text_to_tree():
    from ...docs.snippets.envs import text_to_tree  # noqa


def test_envs_overview_tree_to_list():
    from ...docs.snippets.envs import tree_to_list  # noqa


def test_envs_overview_list_to_observations():
    from ...docs.snippets.envs import lists_to_observations  # noqa
